from typing import List
import os
import datetime
import traceback
import functools
import json
import socket
import requests


class SlackSender(object):

    def __init__(
            self,
            webhook_url: str,
            channel: str,
            user_mentions: List[str] = []):
        super(SlackSender, self).__init__()
        self.webhook_url = webhook_url
        self.channel = channel
        self.user_mentions = user_mentions
        self.dump = {
            "username": "Knock Knock",
            "channel": channel,
            "icon_emoji": ":clapper:",
        }
        self.flag_log = {
            "info": ":flag_green: *COMPLETE!*",
            "warning": ":flag_yellow: *WARNING!*",
            "error": ":flag_red: *ERROR!*",
            "none": ":flag_red: *WRONG LOGGER!* log_type must be info, warning, error"}
        self.flag_start = False
        self.DATE_FORMAT = "%Y-%m-%d %H:%M:%S"

    def requests(self, content):
        now = datetime.datetime.now()
        host_name = socket.gethostname()
        if isinstance(content, dict):
            if 'logger' in content.keys():
                logger = content['logger']
                if log_type not in ['info', 'warning', 'error']:
                    log_type = 'not_found'
                del content['logger']
        contents = {'Status': None,
                    'Starting date': None,
                    'Datetime': None,
                    'Duration': None,
                    'Content': None
                    }
        contenst['status'] = '{} \t {}'.format(
            self.flag_log[log_type], host_name)
        if not self.flag_start:
            self.flag_start = True
            self.start_time = now
            contents['Starting date'] = self.start_time.strftime(
                self.DATE_FORMAT)
        else:
            contents['Datetime'] = self.start_time.strftime(self.DATE_FORMAT)
            elapsed_time = (now - self.start_time).total_seconds()
            self.start_time = now
            elapsed_time = f'{int(elapsed_time//3600)} hours {int(elapsed_time%3600//60)} minutes {int(elapsed_time%3600%60)} seconds'
            contents['`Duration`'] = elapsed_time

        str_value = ''
        if isinstance(content, dict):
            str_value += '\n'
            for key, value in content.items():
                str_value += '\t\t    {:15s}: {}\n'.format(str(key), value)
        else:
            str_value = str(content)
        contents['**Content**'] = str_value
        str_contents = ''
        for key, value in contents.items():
            str_contents += '{}: {}\n'.format(str(key), str(value))

        self.dump['text'] = str_contents
        self.dump['icon_emoji'] = ':tada:'
        requests.post(self.webhook_url, json.dumps(self.dump))

    def __call__(self, func):
        def wrapper(*args, **kwargs):
            start_time = datetime.datetime.now()
            host_name = socket.gethostname()
            func_name = func.__name__
            # Handling distributed training edge case.
            # In PyTorch, the launch of `torch.distributed.launch` sets up a RANK environment variable for each process.
            # This can be used to detect the master process.
            # See https://github.com/pytorch/pytorch/blob/master/torch/distributed/launch.py#L211
            # Except for errors, only the master process will send
            # notifications.
            if 'RANK' in os.environ:
                master_process = (int(os.environ['RANK']) == 0)
                host_name += ' - RANK: %s' % os.environ['RANK']
            else:
                master_process = True
            log_type = 'info'
            if master_process and self.flag_start == False:
                self.flag_start = True
                contents = [
                    '{} \t {}'.format(
                        self.flag_log[log_type].format('INFO'),
                        host_name),
                    '\t\t Function call: %s' %
                    func_name,
                    '\t\t Starting date: %s' %
                    start_time.strftime(
                        self.DATE_FORMAT)]
                contents.append(' '.join(self.user_mentions))
                self.dump['text'] = '\n'.join(contents)
                self.dump['icon_emoji'] = ':clapper:'
                requests.post(self.webhook_url, json.dumps(self.dump))
            try:
                content = func(*args, **kwargs)
                if isinstance(content, dict):
                    if 'logger' in content.keys():
                        logger = value['logger']
                        if logger not in ['info', 'warning', 'error']:
                            logger = 'none'
                        del value['logger']
                if master_process:
                    end_time = datetime.datetime.now()
                    elapsed_time = (end_time - start_time).total_seconds()
                    elapsed_time = f'{int(elapsed_time//3600)} hours {int(elapsed_time%3600//60)} minutes {int(elapsed_time%3600%60)} seconds'
                    contents = [
                        "{} \t {}".format(
                            self.flag_log[log_type].format(
                                self.flag_status[log_type]),
                            host_name),
                        '\t\tFunction call: %s' %
                        func_name,
                        '\t\tDatetime: {}\t:arrow_right:\t{}'.format(
                            start_time.strftime(
                                self.DATE_FORMAT),
                            end_time.strftime(
                                self.DATE_FORMAT)),
                        '\t\tDuration: %s' %
                        str(elapsed_time)]

                    try:
                        str_value = ''
                        if isinstance(value, dict):
                            str_value += '\n'
                            for key, value in content.items():
                                str_value += '\t\t    {:15s}: {}\n'.format(
                                    str(key), value)
                        else:
                            str_value = str(content)
                        contents.append(
                            '\t\tMain call returned value: %s' % str_value)
                    except BaseException:
                        contents.append(
                            'Main call returned value: %s' %
                            "ERROR - Couldn't str the returned value.")

                    contents.append(' '.join(self.user_mentions))
                    self.dump['text'] = '\n'.join(contents)
                    self.dump['icon_emoji'] = ':tada:'
                    requests.post(self.webhook_url, json.dumps(self.dump))

                return content

            except Exception as ex:
                end_time = datetime.datetime.now()
                elapsed_time = end_time - start_time
                contents = ["Your training has crashed ☠️",
                            'Machine name: %s' % host_name,
                            'Main call: %s' % func_name,
                            'Starting date: %s' % start_time.strftime(
                                self.DATE_FORMAT),
                            'Crash date: %s' % end_time.strftime(
                                self.DATE_FORMAT),
                            'Crashed training duration: %s\n\n' % str(
                                elapsed_time),
                            "Here's the error:",
                            '%s\n\n' % ex,
                            "Traceback:",
                            '%s' % traceback.format_exc()]
                contents.append(' '.join(self.user_mentions))
                self.dump['text'] = '\n'.join(contents)
                self.dump['icon_emoji'] = ':skull_and_crossbones:'
                requests.post(webhook_url, json.dumps(self.dump))
                raise ex
        return wrapper



import numpy as np
import matplotlib.pyplot as plt
import math


def imshow(*kwarg, title=None, ncols=2, figsize=(10, 10)):
    num_image = len(kwarg)
    if ncols < 1:
        print('WARNING: Columns must greater than 0. Return None')
        return None
    if num_image < 1:
        print('WARNING: Number of image must greater than 0. Return None')
        return None
    f, ax = plt.subplots(math.ceil(num_image/ncols), ncols,
                         figsize=figsize, squeeze=False)
    for idx, img in enumerate(kwarg):
        ax[int(idx/ncols), int(idx % ncols)].imshow(img)
        if title is not None:
            try:
                ax[int(idx/ncols), int(idx % ncols)].set_title(title[idx])
            except IndexError:
                ax[int(idx/ncols), int(idx % ncols)
                   ].set_title('Figure {}'.format(idx))



import pandas as pd 


class MetricTracker:
    def __init__(self, *keys, writer = None):
        self.writer = writer 
        self._data = pd.DataFrame(index=keys, columns=['total', 'counts', 'average'])
        self.reset()
        
    def reset(self):
        for col in self._data.columns:
            self._data[col].values[:] = 0

    def update(self, key, value, n=1):
        if self.writer is not None:
            self.writer.add_scalar(key, value)
        self._data.total[key] += value * n
        self._data.counts[key] += n
        self._data.average[key] = self._data.total[key] / self._data.counts[key]

    def avg(self, key):
        return self._data.average[key]
    
    def result(self):
        return dict(self._data.average)


import importlib
from datetime import datetime


class TensorboardWriter():
    def __init__(self, log_dir, enabled):
        self.writer = None
        self.selected_module = ""

        if enabled:
            log_dir = str(log_dir)

            # Retrieve vizualization writer.
            succeeded = False
            for module in ["torch.utils.tensorboard", "tensorboardX"]:
                try:
                    self.writer = importlib.import_module(module).SummaryWriter(log_dir)
                    succeeded = True
                    break
                except ImportError:
                    succeeded = False
                self.selected_module = module

            if not succeeded:
                message = "Warning: visualization (Tensorboard) is configured to use, but currently not installed on " \
                    "this machine. Please install TensorboardX with 'pip install tensorboardx', upgrade PyTorch to " \
                    "version >= 1.1 to use 'torch.utils.tensorboard' or turn off the option in the 'config.json' file."
                print(message)

        self.step = 0
        self.mode = ''

        self.tb_writer_ftns = {
            'add_scalar', 'add_scalars', 'add_image', 'add_images', 'add_audio',
            'add_text', 'add_histogram', 'add_pr_curve', 'add_embedding', 'add_graph'
        }
        self.tag_mode_exceptions = {'add_histogram', 'add_embedding'}
        self.timer = datetime.now()

    def set_step(self, step, mode='train'):
        self.mode = mode
        self.step = step
        if step == 0:
            self.timer = datetime.now()
        else:
            duration = datetime.now() - self.timer
            self.add_scalar('steps_per_sec', 1 / duration.total_seconds())
            self.timer = datetime.now()

    def __getattr__(self, name):
        """
        If visualization is configured to use:
            return add_data() methods of tensorboard with additional information (step, tag) added.
        Otherwise:
            return a blank function handle that does nothing
        """
        if name in self.tb_writer_ftns:
            add_data = getattr(self.writer, name, None)

            def wrapper(tag, data, *args, **kwargs):
                if add_data is not None:
                    # add mode(train/valid) tag
                    if name not in self.tag_mode_exceptions:
                        tag = '{}/{}'.format(tag, self.mode)
                    add_data(tag, data, self.step, *args, **kwargs)
            return wrapper
        else:
            # default action for returning methods defined in this class, set_step() for instance.
            try:
                attr = object.__getattr__(name)
            except AttributeError:
                raise AttributeError("type object '{}' has no attribute '{}'".format(self.selected_module, name))
            return attr


import torch
import os
import numpy as np
import pandas as pd 
import time 
from torchvision.utils import make_grid
from pytoan.pytorch.utils import TensorboardWriter, MetricTracker
class Learning(object):
    def __init__(self,
            model,
            criterion,
            optimizer,
            scheduler,
            metric_ftns,
            device,
            num_epoch,
            grad_clipping,
            grad_accumulation_steps,
            early_stopping,
            validation_frequency,
            tensorboard,
            checkpoint_dir,
            resume_path):
        self.device, device_ids = self._prepare_device(device)
        self.model = model.to(self.device)
        self.start_epoch = 1
        if resume_path is not None:
            self._resume_checkpoint(resume_path)
        if len(device_ids) > 1:
            self.model = torch.nn.DataParallel(model, device_ids=device_ids)
        self.criterion = criterion
        self.metric_ftns = metric_ftns
        self.optimizer = optimizer
        self.num_epoch = num_epoch 
        self.scheduler = scheduler
        self.grad_clipping = grad_clipping
        self.grad_accumulation_steps = grad_accumulation_steps
        self.early_stopping = early_stopping
        self.validation_frequency =validation_frequency
        self.checkpoint_dir = checkpoint_dir
        self.best_epoch = 1
        self.best_score = 0
        self.writer = TensorboardWriter(os.path.join(checkpoint_dir, 'tensorboard'), tensorboard)
        self.train_metrics = MetricTracker('loss', writer = self.writer)
        self.valid_metrics = MetricTracker('loss', *[m.__name__ for m in self.metric_ftns], writer = self.writer)
        
    def train(self, train_dataloader, valid_dataloader):
        for epoch in range(self.start_epoch, self.num_epoch+1):
            print("{} epoch: \t start training....".format(epoch))
            start = time.time()
            train_result  = self._train_epoch(epoch, train_dataloader)
            train_result.update({'time': time.time()-start})
            
            for key, value in train_result.items():
                print('    {:15s}: {}'.format(str(key), value))

            if (epoch+1) % self.validation_frequency!=0:
                print("skip validation....")
                continue
            print('{} epoch: \t start validation....'.format(epoch))
            start = time.time()
            valid_result = self._valid_epoch(epoch, valid_dataloader)
            valid_result.update({'time': time.time() - start})
            score = -1
            for key, value in valid_result.items():
                if 'score' in key:
                    score = value 
                print('   {:15s}: {}'.format(str(key), value))
            
            self.post_processing(score, epoch)
            if epoch - self.best_epoch > self.early_stopping:
                print('WARNING: EARLY STOPPING')
                break
    def _train_epoch(self, epoch, data_loader):
        self.model.train()
        self.optimizer.zero_grad()
        self.train_metrics.reset()
        for idx, (data, target) in enumerate(data_loader):
            data, target = data.to(self.device), target.to(self.device)
            output = self.model(data)
            loss = self.criterion(output, target)
            loss.backward()
            self.writer.set_step((epoch - 1) * len(data_loader) + idx)
            self.train_metrics.update('loss', loss.item())
            if (idx+1) % self.grad_accumulation_steps == 0:
                torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.grad_clipping)
                self.optimizer.step()
                self.optimizer.zero_grad()
            if (idx+1) % int(np.sqrt(len(data_loader))) == 0:
                self.writer.add_image('input', make_grid(data.cpu(), nrow=8, normalize=True))
        return self.train_metrics.result()
    def _valid_epoch(self, epoch, data_loader):
        self.valid_metrics.reset()
        self.model.eval()
        with torch.no_grad():
            for idx, (data, target) in enumerate(data_loader):
                data, target = data.to(self.device), target.to(self.device)
                output = self.model(data)
                loss = self.criterion(output, target)
                self.writer.set_step((epoch - 1) * len(data_loader) + idx, 'valid')
                self.valid_metrics.update('loss', loss.item())
                for met in self.metric_ftns:
                    self.valid_metrics.update(met.__name__, met(output, target))
                self.writer.add_image('input', make_grid(data.cpu(), nrow=8, normalize=True))
        
        for name, p in self.model.named_parameters():
            self.writer.add_histogram(name, p, bins='auto')
        
        return self.valid_metrics.result()
    def post_processing(self, score, epoch):
        best = False
        if score > self.best_score:
            self.best_score = score 
            self.best_epoch = epoch 
            best = True
            print("best model: {} epoch - {:.5}".format(epoch, score))
        self._save_checkpoint(epoch = epoch, save_best = best)
        
        if self.scheduler.__class__.__name__ == 'ReduceLROnPlateau':
            self.scheduler.step(score)
        else:
            self.scheduler.step()
    
    def _save_checkpoint(self, epoch, save_best=False):
        """
        Saving checkpoints
        :param epoch: current epoch number
        :param save_best: if True, rename the saved checkpoint to 'model_best.pth'
        """
        arch = type(self.model).__name__
        state = {
            'arch': arch,
            'epoch': epoch,
            'state_dict': self.get_state_dict(self.model),
            'best_score': self.best_score
        }
        filename = os.path.join(self.checkpoint_dir, 'checkpoint_epoch{}.pth'.format(epoch))
        torch.save(state, filename)
        print("Saving checkpoint: {} ...".format(filename))
        if save_best:
            best_path = os.path.join(self.checkpoint_dir, 'model_best.pth')
            torch.save(state, best_path)
            print("Saving current best: model_best.pth ...")
    @staticmethod
    def get_state_dict(model):
        if type(model) == torch.nn.DataParallel:
            state_dict = model.module.state_dict()
        else:
            state_dict = model.state_dict()
        return state_dict
    
    def _resume_checkpoint(self, resume_path):
        resume_path = str(resume_path)
        print("Loading checkpoint: {} ...".format(resume_path))
        checkpoint = torch.load(resume_path, map_location=lambda storage, loc: storage)
        self.start_epoch = checkpoint['epoch'] + 1
        self.best_epoch = checkpoint['epoch']
        self.best_score = checkpoint['best_score']
        self.model.load_state_dict(checkpoint['state_dict'])

        print("Checkpoint loaded. Resume training from epoch {}".format(self.start_epoch))
    
    @staticmethod
    def _prepare_device(device):
        n_gpu_use = len(device)
        n_gpu = torch.cuda.device_count()
        if n_gpu_use > 0 and n_gpu == 0:
            print("Warning: There\'s no GPU available on this machine, training will be performed on CPU.")
            n_gpu_use = 0
        if n_gpu_use > n_gpu:
            print("Warning: The number of GPU\'s configured to use is {}, but only {} are available on this machine.".format(n_gpu_use, n_gpu))
            n_gpu_use = n_gpu
        list_ids = device
        device = torch.device('cuda:{}'.format(device[0]) if n_gpu_use > 0 else 'cpu')
        
        return device, list_ids